/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "VEXPController.h"
#include "controllers/common/family/Families.h"
#include <rapidjson/document.h>
#include <rapidjson/prettywriter.h>
#include <rapidjson/reader.h>
#include <rapidjson/stringbuffer.h>
#include <rapidjson/prettywriter.h>
#include <rapidjson/writer.h>

using namespace cameo;
using namespace std;
using namespace rapidjson;

namespace vexp {

const string VEXPController::TYPE = "vexp_controller";

VEXPController::VEXPController(const string& name) :
	ExperimentController(name) {

	setFamily(family::HIDDEN);

	as.init(this, NOSAVE, "as");
	bs.init(this, NOSAVE, "bs");
	cs.init(this, NOSAVE, "cs");
	aa.init(this, NOSAVE, "aa");
	bb.init(this, NOSAVE, "bb");
	cc.init(this, NOSAVE, "cc");
	ax.init(this, NOSAVE, "ax");
	ay.init(this, NOSAVE, "ay");
	az.init(this, NOSAVE, "az");
	bx.init(this, NOSAVE, "bx");
	by.init(this, NOSAVE, "by");
	bz.init(this, NOSAVE, "bz");
	u.init(this, NOSAVE, "Umatrix");
	b.init(this, NOSAVE, "Bmatrix");

	qh.init(this, NOSAVE, "qh");
	qk.init(this, NOSAVE, "qk");
	ql.init(this, NOSAVE, "ql");
	en.init(this, NOSAVE, "en");
	qm.init(this, NOSAVE, "qm");
	fx.init(this, NOSAVE, "fx");
	ss.init(this, NOSAVE, "ss");

	ki.init(this, NOSAVE, "ki");
	kf.init(this, NOSAVE, "kf");

	a1.init(this, NOSAVE, "a1");
	a2.init(this, NOSAVE, "a2");
	a3.init(this, NOSAVE, "a3");
	a3p.init(this, NOSAVE, "a3p");
	a4.init(this, NOSAVE, "a4");
	a5.init(this, NOSAVE, "a5");
	a6.init(this, NOSAVE, "a6");

	calcvalues.init(this, NOSAVE, "calcvalues");

	calca1.init(this, NOSAVE, "calca1");
	calca2.init(this, NOSAVE, "calca2");
	calca3.init(this, NOSAVE, "calca3");
	calca3p.init(this, NOSAVE, "calca3p");
	calca4.init(this, NOSAVE, "calca4");
	calca5.init(this, NOSAVE, "calca5");
	calca6.init(this, NOSAVE, "calca6");

	calculationStatus.init(this, NOSAVE, "calculation_status");
	calculationError.init(this, NOSAVE, "calculation_error");

	errorMessage.init(this, NOSAVE, "error_message");
	moveModel.init(this, SAVE, "move_model");

	spyPeriod.init(this, SAVE, "spy_period");

	qhScan.init(this, NOSAVE, "qh_scan");
	qkScan.init(this, NOSAVE, "qk_scan");
	qlScan.init(this, NOSAVE, "ql_scan");
	enScan.init(this, NOSAVE, "en_scan");

	dqhScan.init(this, NOSAVE, "dqh_scan");
	dqkScan.init(this, NOSAVE, "dqk_scan");
	dqlScan.init(this, NOSAVE, "dql_scan");
	denScan.init(this, NOSAVE, "den_scan");

	kiScan.init(this, NOSAVE, "ki_scan");
	kfScan.init(this, NOSAVE, "kf_scan");
	fxScan.init(this, NOSAVE, "fx_scan");
	ssScan.init(this, NOSAVE, "ss_scan");

	nbPointsScan.init(this, NOSAVE, "nb_points_scan");
	runningScan.init(this, NOSAVE, "running_scan");

	tasSettings.init(this, "tas_settings");
	sample.init(this, "sample");
	scattering.init(this, "scattering");
	incidentBeam.init(this, "incident_beam");
	scatteredBeam.init(this, "scattered_beam");

	a1Controller.init(this, "A1");
	a2Controller.init(this, "A2");
	a3Controller.init(this, "A3");
	a3pController.init(this, "A3P");
	a4Controller.init(this, "A4");
	a5Controller.init(this, "A5");
	a6Controller.init(this, "A6");

	scanController.init(this, "scan");
}

VEXPController::VEXPController(const VEXPController& controller) :
	ExperimentController(controller) {
}

VEXPController::~VEXPController() {

	m_running.store(false);
	if (m_periodicUpdateThread.get() != nullptr) {
		m_periodicUpdateThread->join();
	}
}

void VEXPController::updateUMatrix() {
	u.update(sample->uRef());
	u.setSize(sample->uRef.getSize());
	u.sendEvent();
}

void VEXPController::updateBMatrix() {
	b.update(tasSettings->b());
	b.setSize(tasSettings->b.getSize());
	b.sendEvent();
}

void VEXPController::refreshCalcParameters(const std::string& value) {

	//cout << "refreshCalcParameters " << value << endl;

	// Parse the values.
	rapidjson::Document document;
	document.Parse(value.c_str());

	rapidjson::Value& ki = document["ki"];
	rapidjson::Value& kf = document["kf"];
	rapidjson::Value& qh = document["qh"];
	rapidjson::Value& qk = document["qk"];
	rapidjson::Value& ql = document["ql"];
	rapidjson::Value& fx = document["fx"];
	rapidjson::Value& ss = document["ss"];

	// Calculate the angles and the status.
	calculate(ki.GetDouble(), kf.GetDouble(), qh.GetDouble(), qk.GetDouble(), ql.GetDouble(), fx.GetInt(), ss.GetInt());
}

double toDouble(double value) {

	if (isinf(value)) {
		return numeric_limits<double>::max();
	}

	if (isnan(value)) {
		return 0.0;
	}

	return value;
}

void VEXPController::refreshFloat64Property(SimpleProperty<float64>& property, float64 value) {

	// Update the value.
	property = value;
}

void VEXPController::refreshFloat64PropertyToMoveModel(SimpleProperty<float64>& property, float64 value) {

	// Update the value only in case moveModel is true.
	if (moveModel()) {
		property = value;
	}
}

void VEXPController::updateScanReference(int index) {

	if (index == 0) {
		updateFloat64PropertyValue(qhScan, scanController->reference.get(0));
	}
	else if (index == 1) {
		updateFloat64PropertyValue(qkScan, scanController->reference.get(1));
	}
	else if (index == 2) {
		updateFloat64PropertyValue(qlScan, scanController->reference.get(2));
	}
	else if (index == 3) {
		updateFloat64PropertyValue(enScan, scanController->reference.get(3));
	}
}

void VEXPController::updateScanDelta(int index) {

	if (index == 0) {
		updateFloat64PropertyValue(dqhScan, scanController->delta.get(0));
	}
	else if (index == 1) {
		updateFloat64PropertyValue(dqkScan, scanController->delta.get(1));
	}
	else if (index == 2) {
		updateFloat64PropertyValue(dqlScan, scanController->delta.get(2));
	}
	else if (index == 3) {
		updateFloat64PropertyValue(denScan, scanController->delta.get(3));
	}
}

void VEXPController::updateNbPoints(int index) {
	updateInt32PropertyValue(nbPointsScan, scanController->nSteps.get(0));
}

void VEXPController::updateScanStatus() {

	if (isHKLScan()) {
		bool running = scanController->commandStatus.isRunning();
		runningScan.update(running);

		if (running) {
			// Get ki, kf, fixed, ss when the scan is starting.
			updateFloat64PropertyValue(kiScan, ki());
			updateFloat64PropertyValue(kfScan, kf());
			updateInt32PropertyValue(fxScan, fx());
			updateInt32PropertyValue(ssScan, ss());
		}
	}
	else {
		runningScan.update(false);
	}
}

void VEXPController::updateProperties() {

	// Copy the values.
	updateFloat64PropertyValue(as, tasSettings->as());
	updateFloat64PropertyValue(bs, tasSettings->bs());
	updateFloat64PropertyValue(cs, tasSettings->cs());
	updateFloat64PropertyValue(aa, tasSettings->aa());
	updateFloat64PropertyValue(bb, tasSettings->bb());
	updateFloat64PropertyValue(cc, tasSettings->cc());
	updateFloat64PropertyValue(ax, tasSettings->ax());
	updateFloat64PropertyValue(ay, tasSettings->ay());
	updateFloat64PropertyValue(az, tasSettings->az());
	updateFloat64PropertyValue(bx, tasSettings->bx());
	updateFloat64PropertyValue(by, tasSettings->by());
	updateFloat64PropertyValue(bz, tasSettings->bz());

	updateUMatrix();
	updateBMatrix();

	updateFloat64PropertyValue(qh, scattering->qh());
	updateFloat64PropertyValue(qk, scattering->qk());
	updateFloat64PropertyValue(ql, scattering->ql());
	updateFloat64PropertyValue(en, scattering->en());
	updateFloat64PropertyValue(qm, scattering->qm());
	updateInt32PropertyValue(fx, scattering->fx());
	updateInt32PropertyValue(ss, scattering->sens());

	updateFloat64PropertyValue(ki, incidentBeam->ki());
	updateFloat64PropertyValue(kf, scatteredBeam->kf());

	a1.update(a1Controller->position());
	a2.update(a2Controller->position());
	a3.update(a3Controller->position());
	a3p.update(a3pController->position());
	a4.update(a4Controller->position());
	a5.update(a5Controller->position());
	a6.update(a6Controller->position());

	updateFloat64PropertyValue(qhScan, scanController->reference.get(0));
	updateFloat64PropertyValue(qkScan, scanController->reference.get(1));
	updateFloat64PropertyValue(qlScan, scanController->reference.get(2));
	updateFloat64PropertyValue(enScan, scanController->reference.get(3));

	updateFloat64PropertyValue(dqhScan, scanController->delta.get(0));
	updateFloat64PropertyValue(dqkScan, scanController->delta.get(1));
	updateFloat64PropertyValue(dqlScan, scanController->delta.get(2));
	updateFloat64PropertyValue(denScan, scanController->delta.get(3));

	updateInt32PropertyValue(nbPointsScan, scanController->nSteps.get(0));

	if (isHKLScan()) {
		runningScan.update(scanController->commandStatus.isRunning());
	}
	else {
		runningScan.update(false);
	}
}

void VEXPController::updateFloat64PropertyValue(SimpleProperty<float64>& property, float64 value) {
	property.update(value);
}

void VEXPController::updateFloat64Property(SimpleProperty<float64>& property, SimpleProperty<float64>& sourceProperty) {
	updateFloat64PropertyValue(property, sourceProperty());
}

void VEXPController::updateInt32PropertyValue(SimpleProperty<int32>& property, int32 value) {
	property.update(value);
}

void VEXPController::updateInt32Property(SimpleProperty<int32>& property, SimpleProperty<int32>& sourceProperty) {
	updateInt32PropertyValue(property, sourceProperty());
}

void VEXPController::registerFloat64PropertyUpdater(SimpleProperty<float64>& property, SimpleProperty<float64>& targetProperty) {
	registerUpdater(property, &VEXPController::updateFloat64Property, this, targetProperty, property);
}

void VEXPController::registerInt32PropertyUpdater(SimpleProperty<int32>& property, SimpleProperty<int32>& targetProperty) {
	registerUpdater(property, &VEXPController::updateInt32Property, this, targetProperty, property);
}

void VEXPController::postConfiguration() {

	// Register the property updaters.
	registerFloat64PropertyUpdater(tasSettings->as, as);
	registerFloat64PropertyUpdater(tasSettings->bs, bs);
	registerFloat64PropertyUpdater(tasSettings->cs, cs);
	registerFloat64PropertyUpdater(tasSettings->aa, aa);
	registerFloat64PropertyUpdater(tasSettings->bb, bb);
	registerFloat64PropertyUpdater(tasSettings->cc, cc);
	registerFloat64PropertyUpdater(tasSettings->ax, ax);
	registerFloat64PropertyUpdater(tasSettings->ay, ay);
	registerFloat64PropertyUpdater(tasSettings->az, az);
	registerFloat64PropertyUpdater(tasSettings->bx, bx);
	registerFloat64PropertyUpdater(tasSettings->by, by);
	registerFloat64PropertyUpdater(tasSettings->bz, bz);

	registerUpdater(sample->uRef, &VEXPController::updateUMatrix, this);
	registerUpdater(sample->b, &VEXPController::updateBMatrix, this);

	registerFloat64PropertyUpdater(scattering->qh, qh);
	registerFloat64PropertyUpdater(scattering->qk, qk);
	registerFloat64PropertyUpdater(scattering->ql, ql);
	registerFloat64PropertyUpdater(scattering->en, en);
	registerFloat64PropertyUpdater(scattering->qm, qm);
	registerInt32PropertyUpdater(scattering->fx, fx);
	registerInt32PropertyUpdater(scattering->sens, ss);

	registerFloat64PropertyUpdater(incidentBeam->ki, ki);
	registerFloat64PropertyUpdater(scatteredBeam->kf, kf);

	registerPropertyCopierByUpdate(a1Controller->position, a1);
	registerPropertyCopierByUpdate(a2Controller->position, a2);
	registerPropertyCopierByUpdate(a3Controller->position, a3);
	registerPropertyCopierByUpdate(a3pController->position, a3p);
	registerPropertyCopierByUpdate(a4Controller->position, a4);
	registerPropertyCopierByUpdate(a5Controller->position, a5);
	registerPropertyCopierByUpdate(a6Controller->position, a6);


	// To calculate the angles.
	registerRefresher(calcvalues, &VEXPController::refreshCalcParameters, this);

	// To move the model.
	registerRefresher(calca1, &VEXPController::refreshFloat64PropertyToMoveModel, this, a1Controller->position);
	registerRefresher(calca2, &VEXPController::refreshFloat64PropertyToMoveModel, this, a2Controller->position);
	registerRefresher(calca3, &VEXPController::refreshFloat64PropertyToMoveModel, this, a3Controller->position);
	registerRefresher(calca3p, &VEXPController::refreshFloat64PropertyToMoveModel, this, a3pController->position);
	registerRefresher(calca4, &VEXPController::refreshFloat64PropertyToMoveModel, this, a4Controller->position);
	registerRefresher(calca5, &VEXPController::refreshFloat64PropertyToMoveModel, this, a5Controller->position);
	registerRefresher(calca6, &VEXPController::refreshFloat64PropertyToMoveModel, this, a6Controller->position);

	// Scan properties.
	registerUpdater(scanController->reference, &VEXPController::updateScanReference, this);
	registerUpdater(scanController->delta, &VEXPController::updateScanDelta, this);
	registerUpdater(scanController->nSteps, &VEXPController::updateNbPoints, this);
	registerStatus(scanController, &VEXPController::updateScanStatus, this);

	updateProperties();

	m_spyPublisher = application::Publisher::create("vexp_spy");

	m_running.store(true);
	m_periodicUpdateThread.reset(new std::thread(std::bind(&VEXPController::periodicUpdate, this)));
}

void VEXPController::calculate(float64 ki, float64 kf, float64 qh, float64 qk, float64 ql, int32 fx, int32 ss) {

	// Reset status.
	calculationStatus.setIdle();
	calculationError = 0;

	try {
		// Ki.
		incidentBeam->calcki = ki;

		// calca1, calca2, calcStatus are calculated.
		incidentBeam->calculate(tas::Beam::K_MODE);

		calca1 = toDouble(incidentBeam->calca1());
		calca2 = toDouble(incidentBeam->calca2());

		// Manage the status.
		if (incidentBeam->calcStatus.isError()) {
			calculationStatus.setError();
			calculationError = incidentBeam->calcError();
		}
		else if (incidentBeam->calcStatus.isWarning()) {
			calculationStatus.setWarning();
		}

		//cout << "Ki -> status " << incidentBeam->calcStatus() << endl;

		// Kf.
		scatteredBeam->calckf = kf;

		// calca1, calca2, calcStatus are calculated.
		scatteredBeam->calculate(tas::Beam::K_MODE);

		calca5 = toDouble(scatteredBeam->calca5());
		calca6 = toDouble(scatteredBeam->calca6());

		// Manage the status.
		if (scatteredBeam->calcStatus.isError()) {
			calculationStatus.setError();
			calculationError = scatteredBeam->calcError();
		}
		else if (scatteredBeam->calcStatus.isWarning() && !calculationStatus.isError()) {
			calculationStatus.setWarning();
		}

		//cout << "Kf -> status " << scatteredBeam->calcStatus() << endl;

		// Q.
		scattering->calcqh = qh;
		scattering->calcqk = qk;
		scattering->calcql = ql;
		scattering->calcen = 2.072 * (ki * ki - kf * kf);
		scattering->calcki = ki;
		scattering->calckf = kf;
		scattering->calcsens = ss;

		// Calculate by fixing Ki.
		scattering->calculate(tas::Beam::K_MODE, fx);

		calca3 = toDouble(scattering->calca3());
		calca3p = toDouble(scattering->calca3p());
		calca4 = toDouble(scattering->calca4());

//		cout << "calc A1 " << calca1() << endl;
//		cout << "calc A2 " << calca2() << endl;
//		cout << "calc A3 " << calca3() << endl;
//		cout << "calc A3P " << calca3p() << endl;
//		cout << "calc A4 " << calca4() << endl;
//		cout << "calc A5 " << calca5() << endl;
//		cout << "calc A6 " << calca6() << endl << endl;


		// Manage the status.
		if (scattering->calcStatus.isError()) {
			calculationStatus.setError();
			calculationError = scattering->calcError();
		}
		else if (scattering->calcStatus.isWarning() && !calculationStatus.isError()) {
			calculationStatus.setWarning();
			cerr << "Warning: Calculated Kf are different" << endl;
		}
	}
	catch (...) {
		calculationStatus.setError();
		calculationError = 0;
	}

//	if (calculationError() != 0) {
//		cout << "VEXP error " << calculationError() << endl;
//	}
}

bool VEXPController::isHKLScan() {

	if (scanController->controller.get(0) != "Scattering") {
		return false;
	}

	return true;
}

std::string VEXPController::getSpyData() {

	PrettyWriter<StringBuffer> writer(s);
	s.Clear();

	writer.StartObject();

	writer.Key("as"); writer.Double(as());
	writer.Key("bs"); writer.Double(bs());
	writer.Key("cs"); writer.Double(cs());
	writer.Key("aa"); writer.Double(aa());
	writer.Key("bb"); writer.Double(bb());
	writer.Key("cc"); writer.Double(cc());
	writer.Key("ax"); writer.Double(ax());
	writer.Key("ay"); writer.Double(ay());
	writer.Key("az"); writer.Double(az());
	writer.Key("bx"); writer.Double(bx());
	writer.Key("by"); writer.Double(by());
	writer.Key("bz"); writer.Double(bz());
	writer.Key("qh"); writer.Double(qh());
	writer.Key("qk"); writer.Double(qk());
	writer.Key("ql"); writer.Double(ql());
	writer.Key("en"); writer.Double(en());
	writer.Key("qm"); writer.Double(qm());

	writer.Key("fx"); writer.Int(fx());
	writer.Key("ss"); writer.Int(ss());

	writer.Key("ki"); writer.Double(ki());
	writer.Key("kf"); writer.Double(kf());

	writer.Key("qhScan"); writer.Double(qhScan());
	writer.Key("qkScan"); writer.Double(qkScan());
	writer.Key("qlScan"); writer.Double(qlScan());
	writer.Key("enScan"); writer.Double(enScan());

	writer.Key("dqhScan"); writer.Double(dqhScan());
	writer.Key("dqkScan"); writer.Double(dqkScan());
	writer.Key("dqlScan"); writer.Double(dqlScan());
	writer.Key("denScan"); writer.Double(denScan());

	writer.Key("nbPointsScan"); writer.Int(nbPointsScan());

	writer.Key("kiScan"); writer.Double(kiScan());
	writer.Key("kfScan"); writer.Double(kfScan());

	writer.Key("fxScan"); writer.Int(fxScan());
	writer.Key("ssScan"); writer.Int(ssScan());

	writer.Key("runningScan"); writer.Bool(runningScan());
	writer.Key("nameScan"); writer.Int(scanController->numor());

	writer.EndObject();

	return s.GetString();
}

void VEXPController::periodicUpdate() {

	while (m_running.load()) {

		std::this_thread::yield();
		std::this_thread::sleep_for(std::chrono::seconds(spyPeriod()));

		std::lock_guard<std::mutex> lock(m_publisherMutex);

		// Send using the publisher.
		string spyData = getSpyData();
		m_spyPublisher->send(spyData);
	}
}

}
