/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef VEXP_VEXPCONTROLLER_H
#define VEXP_VEXPCONTROLLER_H

#include "controllers/tas/common/TasSettings.h"
#include "controllers/tas/common/Scattering.h"
#include "controllers/tas/common/Sample.h"
#include "controllers/tas/common/IncidentBeam.h"
#include "controllers/tas/common/TasScatteredBeam.h"
#include "controllers/common/scanlegacy/ParameterizableScan1D.h"
#include <cameo/cameo.h>
#include <rapidjson/stringbuffer.h>
#include <memory>
#include <unordered_map>
#include <thread>
#include <atomic>

namespace vexp {

/**
 * The VEXPController is the bridge between Nomad and the vEXP application.
 * It groups the properties needed by vEXP:
 * - Sample properties: as, bs, ... U, B matrices.
 * - Scattering properties: qh, ... qm, ki, kf.
 * - Angle properties: A1...A6.
 * These properties are synchronized with the Nomad controllers.
 *
 * The vEXP application is reading these properties when it synchronizes with Nomad (startup and resync).
 * When the vEXP application requests a calculation of the angles, it sends the property calcvalues.
 * Then the calculation is triggered and calca1...calca6 are calculated by the other Nomad controllers.
 */
class VEXPController : public ExperimentController {

public:
	//! Type of controller
	static const std::string TYPE;

	VEXPController(const std::string& name);
	VEXPController(const VEXPController& controller);
	virtual ~VEXPController();

	void updateUMatrix();
	void updateBMatrix();

	void refreshCalcParameters(const std::string& value);
	void refreshFloat64Property(SimpleProperty<float64>& property, float64 value);
	void refreshFloat64PropertyToMoveModel(SimpleProperty<float64>& property, float64 value);

	void updateProperties();

	void updateFloat64PropertyValue(SimpleProperty<float64>& property, float64 value);
	void updateFloat64Property(SimpleProperty<float64>& property, SimpleProperty<float64>& sourceProperty);
	void updateInt32PropertyValue(SimpleProperty<int32>& property, int32 value);
	void updateInt32Property(SimpleProperty<int32>& property, SimpleProperty<int32>& sourceProperty);

	void registerFloat64PropertyUpdater(SimpleProperty<float64>& property, SimpleProperty<float64>& targetProperty);
	void registerInt32PropertyUpdater(SimpleProperty<int32>& property, SimpleProperty<int32>& targetProperty);

	void updateScanReference(int index);
	void updateScanDelta(int index);
	void updateNbPoints(int index);
	void updateScanStatus();

	virtual void postConfiguration();

private:
	void calculate(float64 ki, float64 kf, float64 qh, float64 qk, float64 ql, int32 fx, int32 ss);

	void periodicUpdate();
	std::string getSpyData();

	bool isHKLScan();

	Property<float64> as;
	Property<float64> bs;
	Property<float64> cs;
	Property<float64> aa;
	Property<float64> bb;
	Property<float64> cc;
	Property<float64> ax;
	Property<float64> ay;
	Property<float64> az;
	Property<float64> bx;
	Property<float64> by;
	Property<float64> bz;

	ArrayProperty<float64> u;
	ArrayProperty<float64> b;

	// Real properties.
	Property<float64> qh;
	Property<float64> qk;
	Property<float64> ql;
	Property<float64> en;
	Property<float64> qm;
	Property<int32> fx;
	Property<int32> ss;

	Property<float64> ki;
	Property<float64> kf;

	Property<float64> a1;
	Property<float64> a2;
	Property<float64> a3;
	Property<float64> a3p;
	Property<float64> a4;
	Property<float64> a5;
	Property<float64> a6;

	// JSON String property containing the calculated properties. Use it or the direct properties.
	Property<std::string> calcvalues;

	Property<float64> calca1;
	Property<float64> calca2;
	Property<float64> calca3;
	Property<float64> calca3p;
	Property<float64> calca4;
	Property<float64> calca5;
	Property<float64> calca6;

	StatusProperty calculationStatus;
	Property<int32> calculationError;
	Property<std::string> errorMessage;
	Property<bool> moveModel;

	// Real scan properties.
	Property<float64> qhScan;
	Property<float64> qkScan;
	Property<float64> qlScan;
	Property<float64> enScan;

	Property<float64> dqhScan;
	Property<float64> dqkScan;
	Property<float64> dqlScan;
	Property<float64> denScan;

	Property<float64> kiScan;
	Property<float64> kfScan;
	Property<int32> fxScan;
	Property<int32> ssScan;

	Property<int32> nbPointsScan;
	Property<bool> runningScan;

	// Spy.
	Property<int32> spyPeriod;
	rapidjson::StringBuffer s;

	std::atomic<bool> m_running;
	std::unique_ptr<cameo::application::Publisher> m_spyPublisher;
	std::mutex m_publisherMutex;
	std::unique_ptr<std::thread> m_periodicUpdateThread;

	// Controllers.
	ControllerPtr<tas::TasSettings> tasSettings;
	ControllerPtr<tas::Sample> sample;
	ControllerPtr<tas::Scattering> scattering;
	ControllerPtr<tas::IncidentBeam> incidentBeam;
	ControllerPtr<tas::TasScatteredBeam> scatteredBeam;

	ControllerPtr<axis::AxisController> a1Controller;
	ControllerPtr<axis::AxisController> a2Controller;
	ControllerPtr<axis::AxisController> a3Controller;
	ControllerPtr<axis::AxisController> a3pController;
	ControllerPtr<axis::AxisController> a4Controller;
	ControllerPtr<axis::AxisController> a5Controller;
	ControllerPtr<axis::AxisController> a6Controller;

	ControllerPtr<scan::ParameterizableScan1D> scanController;
};

}

#endif
