/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "StereoElecDef.h"
#include "StereoElecState.h"
#include "StereoElecDriver.h"
#include "PerfectStereoElecDriver.h"
#include "RealStereoElecDriver.h"
#include "drivers/global/DriversCommands.h"
using namespace std;

namespace stereo_elec {

const std::string StereoElecDriver::TYPE = "stereo_elec";
const string StereoElecDriver::READ_CONFIG_COMMAND = "readConfig";
const string StereoElecDriver::DAQ_CONFIG_COMMAND = "daqConfig";
const string StereoElecDriver::READ_MONITORING_COMMAND = "readMonitoring";
/*
 * Constructor
 */
StereoElecDriver::StereoElecDriver(const string& name) :
		DeviceDriver(name), acquisition::NoTofAcquisitionCommon(name), acquisition::TimeGateCommon(name), acquisition::DetectorCommon(
				name) {

	registerStates(new RealStereoElecDriver(this), new PerfectStereoElecDriver(this), new PerfectStereoElecDriver(this));

	mchipaddress.init(this, SAVE, "mchipaddress");
	triggermacaddress.init(this, SAVE, "triggermacaddress");
	triggeripaddress.init(this, SAVE, "triggeripaddress");
	nslot.init(this, SAVE, "nslot");
	nfeboard.init(this, NOSAVE, "nfeboard");
	nledboard.init(this, NOSAVE, "nledboard");
	boardtype.init(this, NOSAVE, "boardtype");
	boardmacaddress.init(this, NOSAVE, "boardmacaddress");
	boardipaddress.init(this, NOSAVE, "boardipaddress");
	namechannel0.init(this, SAVE, "namechannel0");
	namechannel1.init(this, SAVE, "namechannel1");
	namechannel2.init(this, SAVE, "namechannel2");
	namechannel3.init(this, SAVE, "namechannel3");
	namechannel4.init(this, SAVE, "namechannel4");
	namechannel5.init(this, SAVE, "namechannel5");
	namechannel6.init(this, SAVE, "namechannel6");
	namechannel7.init(this, SAVE, "namechannel7");
	slotnumber.init(this, NOSAVE, "slotnumber");

	isconfigured.init(this, NOSAVE, "isconfigured");

	fe_pos5v.init(this, NOSAVE, "fe_pos5v");
	fe_neg5v.init(this, NOSAVE, "fe_neg5v");
	fe_adc3v3a.init(this, NOSAVE, "fe_adc3v3a");
	fe_adc1v8a.init(this, NOSAVE, "fe_adc1v8a");
	fe_tadc03.init(this, NOSAVE, "fe_tadc03");
	fe_tadc47.init(this, NOSAVE, "fe_tadc47");
	led_dac3v3.init(this, NOSAVE, "led_dac3v3");
	led_amp24v.init(this, NOSAVE, "led_amp24v");

	rate0.init(this, NOSAVE, "rate0");
	rate1.init(this, NOSAVE, "rate1");
	rate2.init(this, NOSAVE, "rate2");
	rate3.init(this, NOSAVE, "rate3");
	rate4.init(this, NOSAVE, "rate4");
	rate5.init(this, NOSAVE, "rate5");
	rate6.init(this, NOSAVE, "rate6");
	rate7.init(this, NOSAVE, "rate7");

	t1accepted.init(this, NOSAVE, "t1accepted");
	t1rejected.init(this, NOSAVE, "t1rejected");
	fiforejected.init(this, NOSAVE, "fiforejected");
	t2accepted.init(this, NOSAVE, "t2accepted");
	runname.init(this, NOSAVE, "runname");

	logstring.init(this, NOSAVE, "logstring");

	/*
	 *  Init the device command list
	 */
	initCommand(driver::INIT_COMMAND);
	initCommand(driver::STATUS_COMMAND);
	initCommand(driver::START_COMMAND);
	initCommand(driver::READ_COMMAND);
	initCommand(driver::SYNCHRONIZE_READ_COMMAND);
	initCommand(driver::STOP_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);
	initCommand(DetectorCommon::CLEAR_COMMAND);
	initCommand(AcquisitionCommon::WRITE_PARAMETERS_COMMAND);
	initCommand(DAQ_CONFIG_COMMAND);
	initCommand(READ_CONFIG_COMMAND);
	initCommand(READ_MONITORING_COMMAND);

	registerFunction(NONE_FUNCTION);
	registerRefresher(boardtype, &StereoElecDriver::refreshBoardTypeProperty, this);
	registerRefresher(nfeboard, &StereoElecDriver::refreshNFEBoardProperty, this);
	registerRefresher(nledboard, &StereoElecDriver::refreshNLEDBoardProperty, this);
	registerObserverCommandMs(driver::READ_COMMAND,4); //in millisecond
	registerObserverCommand(driver::READ_INFOS_COMMAND,50); //multiple de 20 ms (quel que soit la valeur de observerCycleDuration-us)
	registerObserverCommand(driver::STATUS_COMMAND,10); //multiple de 20 ms (quel que soit la valeur de observerCycleDuration-us)
	registerSpyCommand(READ_MONITORING_COMMAND, 12);//multiple de 5 sec


	deviceType = ROOT_DEVICE_TYPE_DEVICE_CONTAINER;

	nslot = NBSLOT;
	cout << "nslot= " << nslot() << endl;
	boardtype.resize(nslot());
	boardmacaddress.resize(nslot());
	boardipaddress.resize(nslot());
	namechannel0.resize(nslot());
	namechannel1.resize(nslot());
	namechannel2.resize(nslot());
	namechannel3.resize(nslot());
	namechannel4.resize(nslot());
	namechannel5.resize(nslot());
	namechannel6.resize(nslot());
	namechannel7.resize(nslot());
	slotnumber.resize(nslot());

	feindex.resize(nslot());
	ledindex.resize(nslot());

	fe_pos5v.resize(NBFEMAX);
	fe_neg5v.resize(NBFEMAX);
	fe_adc3v3a.resize(NBFEMAX);
	fe_adc1v8a.resize(NBFEMAX);
	fe_tadc03.resize(NBFEMAX);
	fe_tadc47.resize(NBFEMAX);
	led_dac3v3.resize(NBLEDMAX);
	led_amp24v.resize(NBLEDMAX);

	rate0.resize(NBFEMAX);
	rate1.resize(NBFEMAX);
	rate2.resize(NBFEMAX);
	rate3.resize(NBFEMAX);
	rate4.resize(NBFEMAX);
	rate5.resize(NBFEMAX);
	rate6.resize(NBFEMAX);
	rate7.resize(NBFEMAX);

	nfeboard.update(0);
	nledboard.update(0);

	for(int i=0;i<nslot();i++) slotnumber.set(i,i);

	DAQParams.FEParams_v.resize(0);
	DAQParams.isFEBoard_v.resize(nslot(), false);
	//DAQParams.isFEBoardEnabled_v.resize(nslot(), false);
	isconfigured=false;
	started=false;

	m_SynchroniseReadActivated = false;
	m_SynchroniseReadTerminated = true;
	m_StopActivated = false;

	nbevent=0;

}


/*
 * Destructor
 */
StereoElecDriver::~StereoElecDriver() {
	// Empty
}


void StereoElecDriver::refreshBoardTypeProperty(int32 index, const string& value) throw (CannotSetValue) {
	string oldvalue = boardtype(index);
  	//cout << "boardtype " << boardtype(index) << endl;

	boardtype.update(index, value);
	if (oldvalue == FEBOARD && value != FEBOARD) {
		nfeboard = nfeboard() - 1;
		feslots.erase(feslots.begin() + feindex.at(index));
		computeFEIndex();
	}
	if (oldvalue == LEDBOARD && value != LEDBOARD) {
		nledboard = nledboard() - 1;
		ledslots.erase(ledslots.begin() + ledindex.at(index));
		computeLEDIndex();
	}
	if (oldvalue != FEBOARD && value == FEBOARD) {
		nfeboard = nfeboard() + 1;
		feslots.push_back(index);
		sort(feslots.begin(), feslots.end());
		computeFEIndex();
	}
	if (oldvalue != LEDBOARD && value == LEDBOARD) {
		nledboard = nledboard() + 1;
		ledslots.push_back(index);
		sort(ledslots.begin(), ledslots.end());
		computeLEDIndex();
	}

	cout << "RefreshBoardType, FE at slots : ";
	for (int i = 0; i < feslots.size(); i++) {
		cout << feslots.at(i);
	}
	cout << endl;
	isconfigured=false;

}
void StereoElecDriver::refreshNFEBoardProperty(int32 value) throw (CannotSetValue) {

}

void StereoElecDriver::refreshNLEDBoardProperty(int32 value) throw (CannotSetValue)  {
	if(nledboard()==0) DAQParams.ledBoardInserted=false;
	else DAQParams.ledBoardInserted=true;
	isconfigured=false;
}

/*
 * execute
 */
void StereoElecDriver::execute(const string& aCommand) {

	if ((aCommand != driver::STOP_COMMAND) && (aCommand != driver::STATUS_COMMAND) && (aCommand != driver::READ_COMMAND) && (aCommand != driver::READ_INFOS_COMMAND)) {
		commandProgression = PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER;
	}
	StereoElecState* currentState = dynamic_cast<StereoElecState*>(getCurrentState());

	// Check command
	if (aCommand == driver::INIT_COMMAND) {
		currentState->init();
		currentState->config();
		while(!isconfigured()) sleep(1);

	} else if (aCommand == driver::STATUS_COMMAND) {
		// Status command
	//	cout << "Status command" << endl;
		currentState->readStatus();
	} else if (aCommand == driver::START_COMMAND) {
		// Start command
		//m_StopActivated = false;
		//startActivated = true;
		currentState->start();

	} else if (aCommand == driver::READ_COMMAND) {
		// Read command
		currentState->read();
	} else if (aCommand == driver::SYNCHRONIZE_READ_COMMAND) {
		// Read command
		//		TODO m_ListModeActivated = false;
		m_SynchroniseReadActivated = true;
		m_SynchroniseReadTerminated = false;
		currentState->synchroniseRead();
	} else if (aCommand == driver::STOP_COMMAND) {

		// Stop command
		//m_StopActivated = true;
		//startActivated = false;
		currentState->stop();

	} else if (aCommand == driver::READ_INFOS_COMMAND) {
	//	cout << "Read Infos command" << endl;
		// Info command
		currentState->readInfos();
	} else if (aCommand == DetectorCommon::CLEAR_COMMAND) {
		// Clear command
		sum.update(0.);
		//currentState->clear();
	} else if (aCommand == DetectorCommon::REGROUP_COMMAND) {
		// Write mask command
		//currentState->writeMask();
	} else if (aCommand == AcquisitionCommon::WRITE_PARAMETERS_COMMAND) {
		// Write param command
		currentState->writeParam();
	} else if (aCommand == DAQ_CONFIG_COMMAND) {
		currentState->config();
	} else if (aCommand == READ_CONFIG_COMMAND) {
		currentState->readConfig();
	} else if (aCommand == READ_MONITORING_COMMAND) {
		currentState->readMonitoring();
	}

}

void StereoElecDriver::computeFEIndex() {
	int count = 0;
	for (int32 i = 0; i < nslot(); i++) {
		if (boardtype(i) == FEBOARD) {
			feindex.at(i) = count;
			count++;
		} else
			feindex.at(i) = -1;
	}
}
void StereoElecDriver::computeLEDIndex() {
	int count = 0;
	for (int32 i = 0; i < nslot(); i++) {
		if (boardtype(i) == LEDBOARD) {
			ledindex.at(i) = count;
			count++;
		} else
			ledindex.at(i) = -1;
	}
}


/*
 * calculateProgression
 */
void StereoElecDriver::calculateProgression() {
	int32 progression = 0;
	float64 setpoint = 0.0;
	float64 actual = 0.0;

	if (mode() == GateCommon::TIME_MODE) {
		setpoint = time.setpoint();
		actual = time();
	} else if (mode() == GateCommon::COUNTER1_MODE) {
		setpoint = (float64) counts.setpoint();
		actual = (float64) nbevent;
	} else if (mode() == GateCommon::COUNTER2_MODE) {
		setpoint = (float64) counts2.setpoint();
		actual = (float64) nbevent;
	} else if (mode() == GateCommon::COUNTER3_MODE) {
		setpoint = (float64) counts3.setpoint();
		actual = (float64) counts3();
	}
	if (setpoint == 0.0) {
		commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		cout << "Progression 100 because setpoint=0 " << endl;
	} else {
		progression = (int32) (actual / setpoint * PROGRESSION_RATIO_DEVICE_CONTAINER);
		// Be sure that progression > 0%
		if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER)
			progression = 0;
		// Be sure that only Phased status put 100%
		if (progression >= PROGRESSION_END_DEVICE_CONTAINER)
			progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
		commandProgression = progression;
		if(progression==100) cout << "Progression 100 in calculateProgression()" << endl;

	}
}

/*
 * computeGateStatus
 */
void StereoElecDriver::computeGateStatus(int32 value) {
	int32 valueout = 0;
	/*if (value & NO_MASTER) {
	 valueout |= GateCommon::NO_MASTER_STATUS;
	 } else if (value & WAITFOR_START) {
	 valueout |= GateCommon::WAITFORSTART_STATUS;
	 } else if (value & COUNT_PAUSED) {
	 valueout |= GateCommon::PAUSE_MODE_STATUS;
	 } else if (value & EXT_INHIBIT) {
	 valueout |= GateCommon::EXT_INHIBIT_STATUS;
	 } else if (value & COUNT_ON) {
	 valueout |= GateCommon::RUNNING_STATUS;
	 } else if (value & COUNT_END) {
	 valueout |= GateCommon::END_STATUS;
	 }*/
	gateStatus.update(valueout);
}

/*
 * computeDetectorStatus
 */
void StereoElecDriver::computeDetectorStatus(int32 value) {
	int32 valueout = 0;
	detectorStatus.update(valueout);

}

/*
 * computeAcquisitionStatus
 */
void StereoElecDriver::computeAcquisitionStatus(int32 value) {
	int32 valueout = 0;
	//acquisitionStatus.update(valueout);
}

}
